from ._validator import *
from .clone import *
from .commands import sub_command_group
from .core import (
    AbstractParser,
    argument,
    pos_argument,
    var_argument,
    aliased_argument,
    as_argument,
    copy_argument,
    print_help
)
from .types import *



from types import MethodType
from typing import Callable, TypeVar, overload

from ..validator import Validator

__all__ = [
    'dispatch',
    'validator_for'
]

T = TypeVar('T')
Method = TypeVar('Method', bound=MethodType)
Decorator = Callable[[Method], Method]  # decorator


def dispatch(command: str,
             *alias: str,
             group: str = None,
             order: float = 5,
             usage: str = None,
             hidden=False) -> Decorator:
    """
    A decorator that mark a function a dispatch target function.

    All functions decorated in same dispatch group should have save
    function signature (at least for non-default parameters). For example:

    **Example**

    >>> class D(Dispatch):
    ...     @dispatch('A')
    ...     def function_a(self, a, b, c=None):
    ...         pass
    ...     @dispatch('B')
    ...     def function_b(self, a, b, d=None):
    ...         pass
    ...     def run_function(self):
    ...         self.invoke_command(self, 'A', a, b)

    :param command: primary command name
    :param alias: secondary command names
    :param group: command group
    :param order: order of this command shown in the :meth:`~argclz.dispatch.core.Dispatch.build_command_usages()`
    :param usage: usage line of this command shown in the :meth:`~argclz.dispatch.core.Dispatch.build_command_usages()`
    :param hidden: hide this command from :meth:`~argclz.dispatch.core.Dispatch.list_commands()`
    """

    if len(command) == 0:
        raise ValueError('empty command string')

    def _dispatch(f):
        from .builder import DispatchCommandBuilder
        DispatchCommandBuilder.of(f).build(command, alias, order, group, usage, hidden)
        return f

    return _dispatch


@overload
def validator_for(arg: str) -> Decorator:
    pass


@overload
def validator_for(arg: str, caster: Callable[[str], T] | Validator) -> Decorator:
    pass


@overload
def validator_for(arg: str, caster: Callable[[str], T], validator: Validator) -> Decorator:
    pass


def validator_for(arg: str, caster=None, validator=None):
    """
    A decorator that do the caster and valudation on dispatch arguments.

    **Example**

    >>> class D(Dispatch):
    ...     @dispatch('cmd')
    ...     @validator_for('a')
    ...     def run_cmd(self, a: int):
    ...         assert isinstance(a, int)

    **Type (caster)**

    The parameter ``caster`` usually can be infered via the annotation of target parameter,
    like the `a: int` in above example.

    :param arg: name of the parameter.
    :param caster: type caster with the signature ``(str) -> T``.
    :param validator: validator with the signatire ``(T) -> bool``.
    """
    if isinstance(caster, Validator) and validator is None:
        caster, validator = None, caster

    def _validator_for(f):
        from .builder import DispatchCommandBuilder
        DispatchCommandBuilder.of(f).validator_for(arg, caster, validator)
        return f

    return _validator_for



from __future__ import annotations

from types import UnionType
from typing import TypeVar, Union, Literal, get_origin, get_args, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from .core import Argument

__all__ = ['caster_by_annotation', 'complete_arg_kwargs']

T = TypeVar('T')


def caster_by_annotation(a_name: str, a_type):
    a_type_ori = get_origin(a_type)
    if a_type == Any:
        return None

    if a_type_ori == Literal:
        from .types import literal_type
        return literal_type(get_args(a_type))

    elif a_type_ori == Union or a_type_ori == UnionType:
        a_type_args = get_args(a_type)
        if len(a_type_args) == 2 and get_origin(a_type_args[0]) is Literal and a_type_args[1] == type(None):
            from .types import literal_type
            return literal_type([*get_args(a_type_args[0]), None])

        from .types import union_type
        return union_type(*get_args(a_type))

    elif a_type_ori is not None and (callable(a_type_ori) or isinstance(a_type_ori, type)):
        return a_type_ori

    elif callable(a_type) or isinstance(a_type, type):
        return a_type

    else:
        raise RuntimeError(f'{a_name} {a_type}')


def complete_arg_kwargs(self: Argument):
    if len(self.options) == 0:  # positional argument
        if 'default' not in self.kwargs:
            self.kwargs.setdefault('nargs', '?')

    if 'type' not in self.kwargs:
        # complete bool argument for action and default
        if self.attr_type == bool:
            _complete_arg_kwargs_for_bool(self)

        if get_origin(self.attr_type) is list:
            self.kwargs.setdefault('action', 'append')
        else:
            self.kwargs.setdefault('action', 'store')

        if self.kwargs['action'] in ('store', 'store_const'):  # value type
            _complete_arg_kwargs_for_value(self)

        elif self.kwargs['action'] in ('append', 'append_const', 'extend'):  # collection type
            _complete_arg_kwargs_for_collection(self)

    if get_origin(self.attr_type) is Literal:
        _complete_arg_kwargs_for_literal(self)


def _complete_arg_kwargs_for_bool(self: Argument):
    assert self.attr_type == bool

    if 'default' not in self.kwargs:
        if 'nargs' in self.kwargs:
            from .types import bool_type
            self.kwargs['type'] = bool_type
            self.kwargs['action'] = 'store'
        else:
            self.kwargs.setdefault('action', 'store_true')
            self.kwargs.setdefault('default', False)

    elif self.kwargs['default']:
        self.kwargs.setdefault('action', 'store_false')
        self.kwargs.setdefault('default', True)

    else:
        self.kwargs.setdefault('action', 'store_true')
        self.kwargs.setdefault('default', False)


def _complete_arg_kwargs_for_value(self: Argument):
    a_type_ori = get_origin(self.attr_type)
    if a_type_ori == Union or a_type_ori == UnionType:
        a_type_args = get_args(self.attr_type)
        if len(a_type_args) == 2 and a_type_args[1] == type(None):
            self.kwargs.setdefault('default', None)

    self.kwargs['type'] = caster_by_annotation(self.attr, self.attr_type)


def _complete_arg_kwargs_for_collection(self: Argument):
    self.kwargs.setdefault('default', get_origin(self.attr_type)())

    a_type_arg = get_args(self.attr_type)  # Coll[T]
    if len(a_type_arg) == 0:
        # XXX what kinds of collection it is?
        raise RuntimeError()
    elif len(a_type_arg) == 1:
        self.kwargs['type'] = caster_by_annotation(self.attr, a_type_arg[0])
    else:
        raise RuntimeError()


def _complete_arg_kwargs_for_literal(self: Argument):
    assert get_origin(self.attr_type) is Literal

    from .types import literal_type
    t: literal_type | None
    if isinstance(t := self.kwargs.get('type', None), literal_type):
        t.set_candidate(self.attr_type)

    literal_values = get_args(self.attr_type)
    literal_values = [it for it in literal_values if isinstance(it, str)]

    if isinstance(t := self.kwargs.get('type', None), literal_type):
        if not t.complete:
            self.kwargs.setdefault('choices', literal_values)
    else:
        self.kwargs.setdefault('choices', literal_values)

    self.kwargs.setdefault('metavar', '|'.join(literal_values))





from types import EllipsisType
from typing import TypeVar, Callable, Union, Literal, get_origin, get_args, Type

__all__ = [
    'literal_value_type',
    'bool_type',
    'try_int_type',
    'try_float_type',
    'int_tuple_type',
    'str_tuple_type',
    'float_tuple_type',
    'tuple_type',
    'list_type',
    'union_type',
    'dict_type',
    'slice_type',
    'literal_type',
]

T = TypeVar('T')


def literal_value_type(arg: str) -> bool | int | float | str:
    """Parse a string into its literal Python value"""
    if arg.upper() == 'TRUE':
        return True
    elif arg.upper() == 'FALSE':
        return False
    try:
        return int(arg)
    except ValueError:
        pass

    try:
        return float(arg)
    except ValueError:
        pass

    return arg


def bool_type(value: str) -> bool:
    """Convert a string to a boolean.

    :param value: the input string to evaluate
    :return: False for ('-', '0', 'f', 'false', 'n', 'no', 'x'), True for ('', '+', '1', 't', 'true', 'yes', 'y')
    :raises ValueError: if the string is not recognized as a boolean
    """
    value = value.lower()
    if value in ('-', '0', 'f', 'false', 'n', 'no', 'x'):
        return False
    elif value in ('+', '1', 't', 'true', 'yes', 'y'):
        return True
    else:
        raise ValueError()


def tuple_type(*value_type: Type[T] | Callable[[str], T] | EllipsisType):
    """Create a caster that splits a comma-separated string into a tuple of typed values.

    :param value_type: converter functions for each tuple position; use EllipsisType to repeat last
    :return: a function that converts a comma-separated string into a typed tuple
    """
    try:
        i = value_type.index(...)
    except ValueError:
        pass
    else:
        if i == 0 or i != len(value_type) - 1:
            raise RuntimeError()

    def _type(arg: str) -> tuple[T, ...]:
        ret = []
        remain = ...
        for i, a in enumerate(arg.split(',')):
            if remain is not ...:
                ret.append(remain(a))
            else:
                t = value_type[i]
                if t is ...:
                    remain = value_type[i - 1]
                    ret.append(remain(a))
                else:
                    ret.append(t(a))

        return tuple(ret)

    return _type


str_tuple_type = tuple_type(str, ...)
"""tuple[str, ...]"""
int_tuple_type = tuple_type(int, ...)
"""tuple[int, ...]"""
float_tuple_type = tuple_type(float, ...)
"""tuple[float, ...]"""


def list_type(value_type: Type[T] | Callable[[str], T] = str, *, split=',', prepend: list[T] = None):
    """Caster which converts a delimited string into a list of typed values

    :param value_type: function to convert each element (default: str)
    :param split: delimiter character (default: ',')
    :param prepend: list of values to prepend when string starts with '+' + split
    :return: function that converts a delimited string into a list
    """

    def _cast(arg: str) -> list[T]:
        if arg.startswith(remove := ('+' + split)) and prepend is not None:
            value = list(map(value_type, arg[len(remove):].split(split)))
            return [*prepend, *value]
        else:
            return list(map(value_type, arg.split(split)))

    return _cast


def union_type(*t: Callable[[str], T]):
    """
    Caster that tries multiple converters in order until one succeeds.

    :param t: converter functions to attempt
    :return: function that returns first successful conversion
    :raises TypeError: if all converters fail
    """
    none_type = type(None)

    def _type(arg: str):
        for _t in t:
            if _t is not none_type:
                try:
                    return _t(arg)
                except (TypeError, ValueError):
                    pass
        raise TypeError

    return _type


def dict_type(value_type: Callable[[str], T] = str, default: dict[str, T] = None):
    """Caster that accumulates key-value pairs from 'key:value' or 'key=value' strings.

    :param value_type: function to convert values (default: str)
    :param default: initial dict to populate (default: new dict)
    :return: function that updates and returns the dict
    """
    if default is None:
        default = {}

    def _type(arg: str) -> dict[str, T]:
        if ':' in arg:
            i = arg.index(':')
            value = arg[i + 1:]
            if value_type is not None:
                value = value_type(value)
            default[arg[:i]] = value
        elif '=' in arg:
            i = arg.index('=')
            value = arg[i + 1:]
            if value_type is not None:
                value = value_type(value)
            default[arg[:i]] = value
        elif value_type is None:
            default[arg] = None
        else:
            default[arg] = value_type("")
        return default

    return _type


def slice_type(arg: str) -> slice:
    """Convert a 'start:end' string into a slice object.

    :param arg: string in 'start:end' format
    :return: slice(start, end)
    :raises ValueError: if format is invalid or parts are not integers
    """
    i = arg.index(':')
    v1 = int(arg[:i])
    v2 = int(arg[i + 1:])
    return slice(v1, v2)


def try_int_type(arg: str) -> Union[int, str, None]:
    """Attempt to convert a string to int, returning original or None on failure.

    :param arg: the input string
    :return: int if parsing succeeds, original string if fails, or None if empty"""
    if len(arg) == 0:
        return None
    try:
        return int(arg)
    except ValueError:
        return arg


def try_float_type(arg: str) -> Union[float, str, None]:
    """Attempt to convert a string to float, returning original or None on failure.

    :param arg: the input string
    :return: float if parsing succeeds, original string if fails, or None if empty"""
    if len(arg) == 0:
        return None
    try:
        return float(arg)
    except ValueError:
        return arg


class literal_type:
    """Caster enforcing membership in a set of string literals with optional prefix matching"""

    def __init__(self, candidate: type[Literal] | tuple[str, ...] = None, *,
                 complete: bool = False):
        self.candidate = None
        self.optional = False
        self.complete = complete

        if candidate is not None:
            self.set_candidate(candidate)

    def set_candidate(self, candidate: type[Literal] | tuple[str, ...], overwrite: bool = False):
        if get_origin(candidate) is Literal:
            candidate = get_args(candidate)

        if overwrite or self.candidate is None:
            self.optional = None in candidate
            if self.optional:
                candidate = [it for it in candidate if it is not None]

            self.candidate = tuple(candidate)

    def __call__(self, arg: str):
        if arg in self.candidate:
            return arg

        if not self.complete:
            raise ValueError

        match [it for it in self.candidate if it.startswith(arg)]:
            case []:
                raise ValueError()
            case [match]:
                return match
            case possible:
                raise ValueError(f'confused {possible}')

    def __str__(self):
        return 'Literal' + ('*' if self.complete else '') + '[' + ', '.join(self.candidate) + ']'

    __repr__ = __str__



import argparse
import inspect
import sys
from typing import Type, TypeVar, overload

from .core import AbstractParser, new_parser, ArgumentParser, set_options, ArgumentParserInterrupt

__all__ = [
    'sub_command_group',
    'new_command_parser',
    'parse_command_args',
    'get_sub_command_group'
]

T = TypeVar('T')
ARGCLZ_SUB_COMMANDS = '__argclz_sub_commands__'


class SubCommandGroup:
    def __init__(self, **kwargs):
        self.attr = None
        self.kwargs = kwargs
        self.sub_parsers: list[SubCommand] = []

    def __set_name__(self, owner, name):
        if hasattr(owner, ARGCLZ_SUB_COMMANDS):
            raise RuntimeError('cannot have multiple sub-commands group')

        self.attr = name
        setattr(owner, ARGCLZ_SUB_COMMANDS, self)

    def __get__(self, instance, owner=None):
        if instance is None:
            return self

        try:
            return instance.__dict__[f'__{self.attr}']
        except KeyError:
            pass

        raise AttributeError(self.attr)

    def __set__(self, instance, value):
        instance.__dict__[f'__{self.attr}'] = value

    def __delete__(self, instance):
        try:
            del instance.__dict__[f'__{self.attr}']
        except KeyError:
            pass

    def add_parser(self, ap: argparse.ArgumentParser):
        sb = ap.add_subparsers(**self.kwargs)
        for command in self.sub_parsers:
            command.add_parser(sb, main=self.attr)

    def __call__(self, command: str):
        def _sub_command(clz: Type[AbstractParser]):
            if not issubclass(clz, AbstractParser):
                raise TypeError()

            self.sub_parsers.append(SubCommand(command, clz))
            return clz

        return _sub_command


class SubCommand:
    def __init__(self, command: str, sub_parser: Type[AbstractParser]):
        self.command = command
        self.sub_parser = sub_parser

    def add_parser(self, sb, main='main'):
        pp = new_parser(self.sub_parser, reset=True)
        pp.set_defaults(**{main: self.sub_parser})

        description = self.sub_parser.DESCRIPTION
        if callable(description):
            description = description()

        sb.add_parser(self.command, help=description, parents=[pp], add_help=False)


@overload
def sub_command_group(*, title: str = ...,
                      description: str = ...,
                      required: bool = ...):
    pass


def sub_command_group(**kwargs):
    """
    Create a sub-commands group.

    >>> class Example(AbstractParser):
    ...     command_group = sub_command_group()
    ...     @command_group('a')
    ...     class SubCommand(AbstractParser):
    ...         ...

    The type of ``sub_command_group()`` as an instance-attribute is ``Type[AbstractParser]|None``, and
    its value is handled by :func:`~argclz.core.set_options()` when paring the command-line arguments.

    **Sub command class**

    When parsing successful (especially when ``parse_only=False`` case), sub-command class
    (e.g. ``SubCommand`` in above example) will be initialized. The ``__init__`` could have two
    different signature, there are

    .. code-block:: python

        def __init__(self): ... # no-arg init
        def __init__(self, parent: AbstractParser): ... # one-arg init

    where the parameter ``parent`` refer to its outer ``AbstractParser`` (e.g. ``Example`` in above example).


    """
    return SubCommandGroup(**kwargs)


def get_sub_command_group(instance) -> SubCommandGroup | None:
    """(internal function)"""
    if not isinstance(instance, type):
        instance = type(instance)

    return getattr(instance, ARGCLZ_SUB_COMMANDS, None)


def init_sub_command(p: AbstractParser) -> AbstractParser:
    """(internal function)"""
    if (sub := get_sub_command_group(p)) is None:
        return p

    try:
        pp = sub.__get__(p)
    except AttributeError:
        return p

    if pp is None:
        return p

    if not isinstance(pp, type):
        return pp

    s = inspect.signature(pp.__init__)
    if len(s.parameters) == 1:
        return pp()
    else:
        return pp(p)


def new_command_parser(parsers: dict[str, AbstractParser | Type[AbstractParser]],
                       usage: str = None,
                       description: str = None,
                       **kwargs) -> ArgumentParser:
    """A convenient way to create an ArgumentParser with sub-commands.

    :param parsers: dict of command to :class:`~argclz.core.AbstractParser`.
    :param usage: parser usage
    :param description: parser description
    :return:
    """
    ap = ArgumentParser(
        usage=usage,
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        **kwargs
    )

    group = SubCommandGroup(title='commands')
    group.attr = 'main'
    group.sub_parsers = [SubCommand(cmd, pp) for cmd, pp in parsers.items()]
    group.add_parser(ap)

    return ap


def parse_command_args(parsers: ArgumentParser | dict[str, AbstractParser | Type[AbstractParser]],
                       args: list[str] = None,
                       usage: str = None,
                       description: str = None,
                       parse_only=False,
                       system_exit: Type[BaseException] = SystemExit) -> AbstractParser:
    """
    A convenient way to run an ArgumentParser with sub-commands.

    :param parsers: dict of command to :class:`~argclz.core.AbstractParser`.
    :param args: List of strings representing the command-line input (e.g. `sys.argv[1:]`). If ``None``, defaults to current process args
    :param usage: Optional usage string to override the auto-generated help
    :param description: Optional description for the main parser
    :param system_exit: exit when commandline parsed fail.
    :return: The parser instance that handled the command (or an :class:`ArgumentParsingResult` if ``parse_only`` or ``system_exit=False``)
    """
    if isinstance(parsers, ArgumentParser):
        parser = parsers
    else:
        parser = new_command_parser(parsers, usage, description)

    try:
        result = parser.parse_args(args)
    except ArgumentParserInterrupt as e:
        exit_status = e.status
        exit_message = e.message
        pp = None
    else:
        exit_status = None
        exit_message = None

        pp: AbstractParser = getattr(result, 'main', None)
        if isinstance(pp, type):
            pp = pp()

        if pp is not None:
            set_options(pp, result)

    if parse_only:
        return pp

    if exit_status is not None:
        if system_exit is SystemExit:
            if exit_status != 0:
                parser.print_usage(sys.stderr)
                print(exit_message, file=sys.stderr)
            sys.exit(exit_status)
        elif issubclass(system_exit, BaseException):
            raise system_exit(exit_status)
        else:
            sys.exit(exit_status)

    if pp is not None:
        pp.run()

    return pp



import inspect
import sys
from typing import Callable, TypeVar, Generic

from .core import ARGCLZ_DISPATCH_COMMAND, DispatchCommand
from ..validator import Validator

__all__ = ['DispatchCommandBuilder']

T = TypeVar('T')
P = inspect.Parameter


class DispatchCommandBuilder:
    """
    (internal) Do not use class directly.
    """
    def __init__(self, func):
        self.func = func
        self.signature = inspect.signature(func, eval_str=True)
        self.validators = {}

    @classmethod
    def of(cls, f):
        ret = getattr(f, ARGCLZ_DISPATCH_COMMAND, None)
        if ret is None:
            ret = DispatchCommandBuilder(f)
            setattr(f, ARGCLZ_DISPATCH_COMMAND, ret)
        elif isinstance(ret, DispatchCommandBuilder):
            pass
        elif isinstance(ret, DispatchCommand):
            raise RuntimeError(f'{f.__name__} already frozen')
        else:
            raise TypeError()
        return ret

    def validator_for(self, arg: str, caster: Callable[[str], T] = None, validator: Validator = None):
        try:
            p = self.signature.parameters[arg]
        except KeyError:
            print(f'unknown arg name : {arg} for function {self.func.__name__}', file=sys.stderr)
            return

        if caster is None:
            from .._types import caster_by_annotation
            if p.annotation is P.empty:
                raise RuntimeError(f'missing type : {self.func.__name__}({arg})')

            caster = caster_by_annotation(arg, p.annotation)

        self.validators[arg] = TypeCasterWithValidator(caster, validator)

    def build(self, command: str,
              aliases: tuple[str, ...],
              order: float = 5,
              group: str = None,
              usage: str = None,
              hidden=False) -> DispatchCommand:
        ret = DispatchCommand(None, group, command, aliases, order, usage, self.func, self.validators, hidden)
        setattr(self.func, ARGCLZ_DISPATCH_COMMAND, ret)
        return ret


class TypeCasterWithValidator(Generic[T]):
    """
    (internal) Do not use class directly.
    """
    def __init__(self, caster: Callable[[str], T] | None,
                 validator: Callable[[T], bool]):
        self.caster = caster
        self.validator = validator

    def __call__(self, value: str) -> T:
        raw_value = value

        if self.caster is not None:
            try:
                value = self.caster(value)
            except BaseException as e:
                # print(e)
                raise

        if self.validator is not None:
            try:
                fail = not self.validator(value)
            except BaseException as e:
                # print(e)
                raise
            else:
                if fail:
                    raise ValueError(f'fail validation : "{raw_value}"')

        return value



from __future__ import annotations

import abc
import argparse
import collections
import sys
from collections.abc import Sequence, Iterable, Callable
from typing import Type, TypeVar, Literal, overload, Any, Optional, get_type_hints, TextIO

from typing_extensions import Self

__all__ = [
    # parser
    'AbstractParser',
    'new_parser',
    'parse_args',
    # argument
    'Argument',
    'argument',
    'pos_argument',
    'var_argument',
    'aliased_argument',
    'as_argument',
    # utilities
    'foreach_arguments',
    'with_defaults',
    'set_options',
    'copy_argument',
    'print_help',
    'as_dict'
]

T = TypeVar('T')
Nargs = Literal[
    '*', '+', '?', '...'
]
Actions = Literal[
    'store',
    'store_const',
    'store_true',
    'store_false',
    'append',
    'append_const',
    'extend',
    'count',
    'help',
    'version',
        #
    'boolean'
]


class ArgumentParserInterrupt(RuntimeError):
    """(internal) Error raised when any error occurs during command-line parsing."""

    def __init__(self, status: int, message: str | None):
        """
        :param status: error code
        :param message: error message
        """
        if message is None:
            super().__init__(f'exit {status}')
        else:
            super().__init__(f'exit {status}: {message}')

        self.status = status
        self.message = message


class ArgumentParser(argparse.ArgumentParser):
    """(internal) override ``argparse.ArgumentParser``."""

    def exit(self, status: int = 0, message: str = None):
        raise ArgumentParserInterrupt(status, message)

    def error(self, message: str):
        raise ArgumentParserInterrupt(2, message)


class AbstractParser(metaclass=abc.ABCMeta):
    """Commandline parser."""

    USAGE: str | list[str] = None
    """parser usage."""

    DESCRIPTION: str = None
    """parser description. Could be override as a method if its content is dynamic-generated."""

    EPILOG: str = None
    """parser epilog. Could be override as a method if its content is dynamic-generated."""

    def __new__(cls, *args, **kwargs):
        obj = object.__new__(cls)
        with_defaults(obj)
        return obj

    @classmethod
    def new_parser(cls, **kwargs) -> ArgumentParser:
        """create an :class:`~argclz.core.ArgumentParser`.

        :param kwargs: keyword parameters to ``argparse.ArgumentParser``
        :return: an ArgumentParser.
        """
        return new_parser(cls, **kwargs)

    def main(self, args: list[str] | None = None, *,
             parse_only=False,
             system_exit: Type[BaseException] = SystemExit) -> AbstractParser:
        """parsing the commandline input *args* and call :meth:`~argclz.core.ArgumentParser.run()`.

        :param args: command-line arguments. If omitted, use ``sys.args``.
        :param parse_only: parse command-line arguments only, do not raise error and invoke :meth:`~argclz.core.ArgumentParser.run()`
        :param system_exit: error raised when commandline parsed fail.
        :return: parser itself. If it has sub command, return sub parser when used.
        """
        parser = self.new_parser(reset=True)

        try:
            result = parser.parse_args(args)
        except ArgumentParserInterrupt as e:
            exit_status = e.status
            exit_message = e.message
            result = None
        else:
            exit_status = None
            exit_message = None
            set_options(self, result)

        from .commands import init_sub_command
        pp = init_sub_command(self)
        if pp is not self and result is not None:
            set_options(pp, result)

        if parse_only:
            return pp

        if exit_status is not None:
            if system_exit is SystemExit:
                if exit_status != 0:
                    parser.print_usage(sys.stderr)
                    print(exit_message, file=sys.stderr)
                sys.exit(exit_status)
            elif issubclass(system_exit, BaseException):
                raise system_exit(exit_status)
            else:
                sys.exit(exit_status)

        pp.run()

        return pp

    def run(self):
        """called after :meth:`~argclz.core.ArgumentParser.main()`"""
        pass

    def __str__(self):
        return type(self).__name__

    def __repr__(self):
        """print key value pair content"""
        self_type = type(self)
        ret = []
        for a_name in dir(self_type):
            a_value = getattr(self_type, a_name)
            if isinstance(a_value, Argument) and not a_name.startswith('_'):
                try:
                    ret.append(f'{a_name} = {a_value.__get__(self, self_type)}')
                except:
                    ret.append(f'{a_name} = <error>')
            elif isinstance(a_value, property):
                try:
                    ret.append(f'{a_name} = {a_value.__get__(self, self_type)}')
                except:
                    ret.append(f'{a_name} = <error>')

        return '\n'.join(ret)


class Argument(object):
    """
    (internal) Do not use class directly.

    Commandline argument descriptor (https://docs.python.org/3/glossary.html#term-descriptor).
    Carried the arguments pass to ``argparse.ArgumentParser.add_argument``.

    """

    def __init__(self, *options,
                 validator: Callable[[T], bool] = None,
                 group: str = None,
                 ex_group: str = None,
                 hidden: bool = False,
                 **kwargs):
        """

        :param options: options
        :param validator: argument validator.
        :param group: argument group.
        :param ex_group: mutually exclusive group.
        :param hidden: hide this argument from help document
        :param kwargs:
        """
        from .validator import Validator
        if len(options) > 0 and isinstance(options[-1], Validator):
            if validator is not None:
                raise RuntimeError()
            validator = options[-1]
            options = options[:-1]

        if not all([it.startswith('-') for it in options]):
            raise RuntimeError(f'options should startswith "-". {options}')

        if isinstance(validator, Validator):
            validator = validator.freeze()

        self.attr = None
        self.attr_type = Any
        self.group = group
        self.ex_group = ex_group
        self.validator = validator
        self.options = options
        self.hidden = hidden
        self._kwargs = kwargs  # original kwargs
        self.kwargs = dict(kwargs)

    @property
    def default(self):
        try:
            return self.kwargs['default']
        except KeyError:
            pass

        if self.attr_type == bool:
            return self.kwargs.get('action', 'store_true') != 'store_true'

        if self.kwargs.get('action', None) in ('append', 'extend', 'append_const'):
            return []

        raise ValueError

    @property
    def const(self):
        try:
            return self.kwargs['const']
        except KeyError:
            pass

        if self.attr_type == bool:
            return self.kwargs.get('action', 'store_true') == 'store_true'

        raise ValueError

    @property
    def metavar(self) -> Optional[str]:
        return self.kwargs.get('metavar', None)

    @property
    def choices(self) -> Optional[tuple[str, ...]]:
        return self.kwargs.get('choices', None)

    @property
    def required(self) -> bool:
        return self.kwargs.get('required', False)

    @property
    def type(self) -> type | Callable[[str], T]:
        try:
            return self.kwargs['type']
        except KeyError:
            pass

        attr_type = self.attr_type
        if attr_type == bool:
            from .types import bool_type
            return bool_type
        elif attr_type in (str, int, float):
            return attr_type
        else:
            from ._types import caster_by_annotation
            return caster_by_annotation(self.attr, attr_type)

    @property
    def help(self) -> Optional[str]:
        return self.kwargs.get('help', None)

    def __set_name__(self, owner: Type, name: str):
        if self.attr is not None:
            raise RuntimeError('reuse Argument')

        self.attr = name
        self.attr_type = get_type_hints(owner).get(name, Any)

        from ._types import complete_arg_kwargs
        complete_arg_kwargs(self)

        if self.hidden:
            self.kwargs['help'] = argparse.SUPPRESS

    def __get__(self, instance, owner=None):
        if instance is None:
            if owner is not None:  # ad-hoc for the document building
                self.__doc__ = self.help
            return self
        try:
            return instance.__dict__[f'__{self.attr}']
        except KeyError:
            pass

        raise AttributeError(self.attr)

    def __set__(self, instance, value):
        if (validator := self.validator) is not None:
            from .validator import Validator, ValidatorFailError
            try:
                fail = not validator(value)
            except ValidatorFailError:
                raise
            except BaseException as e:
                raise ValueError('validator fail') from e
            else:
                if fail:
                    raise ValueError('validator fail')

        instance.__dict__[f'__{self.attr}'] = value

    def __delete__(self, instance):
        try:
            del instance.__dict__[f'__{self.attr}']
        except KeyError:
            pass

    def add_argument(self, ap: argparse._ActionsContainer, instance):
        """Add this into `argparse.ArgumentParser`.

        :param ap:
        :param instance:
        :return:
        """
        try:
            return ap.add_argument(*self.options, **self.kwargs, dest=self.attr)
        except TypeError as e:
            if isinstance(instance, type):
                name = instance.__name__
            else:
                name = type(instance).__name__
            raise RuntimeError(f'{name}.{self.attr} : ' + repr(e)) from e

    @overload
    def with_options(self,
                     option: str | dict[str, str] = None,
                     *options: str,
                     action: Actions = None,
                     nargs: int | Nargs = None,
                     const: T = None,
                     default: T = None,
                     type: Type | Callable[[str], T] = None,
                     validator: Callable[[T], bool] = None,
                     choices: Sequence[str] = None,
                     required: bool = None,
                     hidden: bool = None,
                     help: str = None,
                     group: str = None,
                     metavar: str = None) -> Self:
        pass

    def with_options(self, *options, **kwargs) -> Self:
        """Modify or update keyword parameter and return a new argument.

        option flags update rule:

        1. ``()``, ``(...)`` : do not update options
        2. ``('-a', '-b')`` : replace options
        3. ``(..., '-c')`` : additional options
        4. ``({'-a': '-A'})`` : rename options
        5. ``({'-a': ...})`` : remove options
        6. ``({'-a': '-A', '-b': ...}, '-c')`` : rename '-a', remove '-b', add '-c'

        general form:

            ``() | (dict?, ...?, *str)``

        :param options: change option flags
        :param kwargs: change keyword parameters, use `...` to unset parameter
        :return:
        """
        kw = dict(self._kwargs)  # use original kwargs
        kw['group'] = self.group
        kw['ex_group'] = self.ex_group
        kw['validator'] = self.validator
        kw['hidden'] = self.hidden
        kw.update(kwargs)

        for k in list(kw.keys()):
            if kw[k] is ...:
                del kw[k]

        cls = type(self)

        if len(self.options) > 0:
            match options:
                case ():
                    return cls(*self.options, **kw)
                case (e, *o) if e is ...:
                    return cls(*self.options, *o, **kw)
                case (dict(d), ):
                    return cls(*self._map_options(d), **kw)
                case (dict(d), *o):
                    return cls(*self._map_options(d), *o, **kw)
                case _:
                    return cls(*options, **kw)
        else:
            if len(options) > 0:
                raise RuntimeError('cannot change positional argument to optional')

            return cls(**kw)

    def _map_options(self, mapping: dict[str, str]) -> list[str]:
        ret = []
        for old_opt in self.options:
            try:
                new_opt = mapping[old_opt]
            except KeyError:
                ret.append(old_opt)
            else:
                if new_opt is not ...:
                    ret.append(new_opt)
        return ret


@overload
def argument(*options: str,
             action: Actions = ...,
             nargs: int | Nargs = ...,
             const: T = ...,
             default: T = ...,
             type: Type | Callable[[str], T] = ...,
             validator: Callable[[T], bool] = ...,
             choices: Sequence[str] = ...,
             required: bool = False,
             hidden: bool = False,
             help: str = ...,
             group: str = None,
             ex_group: str = None,
             metavar: str = ...) -> T:
    pass


def argument(*options: str, **kwargs):
    """create an argument attribute

    **Usage**

    >>> class Example:
    ...     a: str = argument('-a')

    **Type (caster)**

    The parameter ``type`` usually can be infered via the annotation of target attribute,
    like the `a: str` in above example.

    There are some special rules to treat common type:

    - **Literal types**: ``Literal[...]`` annotations create a caster that restricts values to the specified literals.
    - **Boolean flags**: annotating with ``bool`` automatically sets up a `store_true` or `store_false` action, making the option a flag.
    - **Container types**: ``list[T]`` annotations infer a repeated option with `append` (or `extend`) action, converting each input to type ``T``.
    - **Tuple types**: tuple annotations (e.g. ``tuple[int, ...]``) use a comma-separated parser to split and convert values into a tuple of ``T``.

    If the default type infered does not fit your application, you can give it directly.

    **Validator**

    Although parameter ``type`` can handle some validation works, it is ignored when user
    assign value to the attribute directly. The parameter ``validator`` is used to perform the assignment
    validation work and raise a ``ValueError`` (normal case) if any improper assignments.

    The parameter ``validator`` has been treated specially when use :attr:`~argclz.validator`, which it
    can put at lastest position of ``options``

    >>> from argclz import argument, validator
    >>> class Example:
    ...     a: str = argument('-a', validator.str.match(r'\d+'))

    :param options: options strings
    :param action: argument action. Please see ``argparse.ArgumentParser.add_argument(action)`` for detailed.
    :param nargs: number of following values. Please see ``argparse.ArgumentParser.add_argument(nargs)`` for detailed.
    :param const: Please see ``argparse.ArgumentParser.add_argument(const)`` for detailed.
    :param default: default value of argument. Please see ``argparse.ArgumentParser.add_argument(default)`` for detailed.
    :param type: type caster with signature ``(str) -> T``. Please see ``argparse.ArgumentParser.add_argument(type)`` for detailed.
    :param validator: value validator with signature ``(T) -> bool``.
    :param choices: Please see ``argparse.ArgumentParser.add_argument(choices)`` for detailed.
    :param required: Please see ``argparse.ArgumentParser.add_argument(required)`` for detailed.
    :param hidden: hide this argument from help document.
    :param help: help document for this argument.
    :param group: group name of this argument.
    :param ex_group: the mutually exclusive group name of this argument.
    :param metavar: name of argument value. Please see ``argparse.ArgumentParser.add_argument(metavar)`` for detailed.
    """
    return Argument(*options, **kwargs)


@overload
def pos_argument(option: str,
                 validator: Callable[[T], bool] = ..., *,
                 nargs: Nargs = None,
                 action: Actions = ...,
                 const=...,
                 default=...,
                 type: Type | Callable[[str], T] = ...,
                 choices: Sequence[str] = ...,
                 required: bool = False,
                 help: str = ..., ) -> T:
    pass


def pos_argument(option: str, validator: Callable[[T], bool] = ..., *, nargs=None, **kwargs):
    """create a positional (non-flag) command-line argument attribute.

    **Usage**

    >>> class Example:
    ...     a: str = pos_argument('A')

    shorten for ``argument(metavar=option, nargs=nargs, validator=validator, **kwargs)``

    :param option: The name for the positional argument shown in usage messages
    :param validator: value validator with signature ``(T) -> bool``.
    :param nargs: number of following values. Please see ``argparse.ArgumentParser.add_argument(nargs)`` for detailed.
    :param action: argument action. Please see ``argparse.ArgumentParser.add_argument(action)`` for detailed.
    :param const: Please see ``argparse.ArgumentParser.add_argument(const)`` for detailed.
    :param default: default value of argument. Please see ``argparse.ArgumentParser.add_argument(default)`` for detailed.
    :param type: type caster with signature ``(str) -> T``. Please see ``argparse.ArgumentParser.add_argument(type)`` for detailed.
    :param choices: Please see ``argparse.ArgumentParser.add_argument(choices)`` for detailed.
    :param required: Please see ``argparse.ArgumentParser.add_argument(required)`` for detailed.
    :param help: help document for this argument.
    """
    if validator is not ...:
        kwargs['validator'] = validator
    return Argument(metavar=option, nargs=nargs, **kwargs)


@overload
def var_argument(option: str,
                 validator: Callable[[T], bool] = ..., *,
                 nargs: Nargs = ...,
                 action: Actions = ...,
                 type: Type | Callable[[str], T] = ...,
                 help: str = ...) -> list[T]:
    pass


def var_argument(option: str, validator: Callable[[T], bool] = ..., *, nargs='*', action='extend', **kwargs):
    """
    create a variable-length positional argument, suitable for capturing multiple values into a list.

    **Usage**

    >>> class Example:
    ...     a: list[str] = var_argument('A')

    shorten for ``argument(metavar=option, nargs='*', action='extend', validator=validator, **kwargs)`` by default.

    :param option: The name for the vary-length positional argument shown in usage messages
    :param validator:  value validator with signature ``(T) -> bool``.
    :param nargs: number of following values. Please see ``argparse.ArgumentParser.add_argument(nargs)`` for detailed.
    :param action: argument action. Please see ``argparse.ArgumentParser.add_argument(action)`` for detailed.
    :param type: type caster with signature ``(str) -> T``. Please see ``argparse.ArgumentParser.add_argument(type)`` for detailed.
    :param help: help document for this argument.
    """
    if validator is not ...:
        kwargs['validator'] = validator
    return Argument(metavar=option, nargs=nargs, action=action, **kwargs)


class AliasArgument(Argument):
    """
    (internal) Do not use class directly.
    """
    def __init__(self, *options,
                 aliases: dict[str, Any],
                 **kwargs):
        super().__init__(*options, **kwargs)
        self.aliases = aliases

    def add_argument(self, ap: argparse._ActionsContainer, owner):
        gp = ap.add_mutually_exclusive_group(required=self.required)
        super().add_argument(gp, owner)

        primary = self.options[0]
        for name, values in self.aliases.items():
            kw = dict(self.kwargs)
            kw.pop('metavar', None)
            kw.pop('type', None)
            kw['action'] = 'store_const'
            kw['const'] = values
            kw['help'] = f'short for {primary}={values}.'
            gp.add_argument(name, **kw, dest=self.attr)


@overload
def aliased_argument(options: str, *,
                     aliases: dict[str, T],
                     nargs: Nargs = ...,
                     action: Actions = ...,
                     const=...,
                     default=...,
                     type: Type | Callable[[str], T] = ...,
                     validator: Callable[[T], bool] = ...,
                     choices: Sequence[str] = ...,
                     help: str = ...,
                     group: str = None,
                     ex_group: str = None,
                     metavar: str = ...) -> T:
    pass


def aliased_argument(*options: str, aliases: dict[str, T], **kwargs):
    """
    create an argument that supports shorthand aliases for specific constant values.

    **Usage**

    >>> class Example:
    ...     level: str = aliased_argument(
    ...         '--level',
    ...         aliases={
    ...             '--low': 'low',
    ...             '--mid': 'middle',
    ...             '--high': 'high',
    ...         },
    ...         choices=('low', 'middle', 'high')
    ...     )

    :param options: options strings
    :param aliases: a dictionary maps options to value.
    :param action: argument action. Please see ``argparse.ArgumentParser.add_argument(action)`` for detailed.
    :param nargs: number of following values. Please see ``argparse.ArgumentParser.add_argument(nargs)`` for detailed.
    :param const: Please see ``argparse.ArgumentParser.add_argument(const)`` for detailed.
    :param default: default value of argument. Please see ``argparse.ArgumentParser.add_argument(default)`` for detailed.
    :param type: type caster with signature ``(str) -> T``. Please see ``argparse.ArgumentParser.add_argument(type)`` for detailed.
    :param validator: value validator with signature ``(T) -> bool``.
    :param choices: Please see ``argparse.ArgumentParser.add_argument(choices)`` for detailed.
    :param required: Please see ``argparse.ArgumentParser.add_argument(required)`` for detailed.
    :param hidden: hide this argument from help document.
    :param help: help document for this argument.
    :param group: group name of this argument.
    :param ex_group: the mutually exclusive group name of this argument.
    :param metavar: name of argument value. Please see ``argparse.ArgumentParser.add_argument(metavar)`` for detailed.
    """
    return AliasArgument(*options, aliases=aliases, **kwargs)


def as_argument(a) -> Argument:
    """cast argument attribute as an :class:`~argclz.core.Argument` for type checking framework/IDE."""
    if isinstance(a, Argument):
        return a
    raise TypeError


def foreach_arguments(instance: T | Type[T]) -> Iterable[Argument]:
    """iterating all argument attributes in instance.

    :param instance: any instance that contains ``argument``.
    :return:
    """
    if isinstance(instance, type):
        clazz = instance
    else:
        clazz = type(instance)

    arg_set = set()
    for clz in reversed(clazz.mro()):
        if (ann := getattr(clz, '__annotations__', None)) is not None:
            for attr in ann:
                if isinstance((arg := getattr(clazz, attr, None)), Argument) and attr not in arg_set:
                    arg_set.add(attr)
                    yield arg


def new_parser(instance: T | Type[T], reset=False, **kwargs) -> ArgumentParser:
    """Create ``ArgumentParser`` for instance.

    :param instance: any instance that contains ``argument``.
    :param reset: reset argument attributes. do nothing if *instance* isn't an instance.
    :param kwargs: Please see ``argparse.ArgumentParser(**kwargs)`` for detailed.
    :return:
    """
    if isinstance(instance, AbstractParser) or (isinstance(instance, type) and issubclass(instance, AbstractParser)):
        usage = instance.USAGE
        if isinstance(usage, list):
            usage = '\n       '.join(usage)
        kwargs.setdefault('usage', usage)

        description = instance.DESCRIPTION
        if callable(description):
            description = description()
        kwargs.setdefault('description', description)

        epilog = instance.EPILOG
        if callable(epilog):
            epilog = epilog()
        kwargs.setdefault('epilog', epilog)

        kwargs.setdefault('formatter_class', argparse.RawTextHelpFormatter)

    ap = ArgumentParser(**kwargs)

    groups: dict[str, list[Argument]] = collections.defaultdict(list)

    from .commands import get_sub_command_group
    if (sub := get_sub_command_group(instance)) is not None:
        sub.add_parser(ap)

    # setup non-grouped arguments
    mu_ex_groups: dict[str, argparse._ActionsContainer] = {}
    for arg in foreach_arguments(instance):
        if instance is not None and not isinstance(instance, type) and reset:
            arg.__delete__(instance)

        if arg.group is not None:
            groups[arg.group].append(arg)
            continue
        elif arg.ex_group is not None:
            try:
                tp = mu_ex_groups[arg.ex_group]
            except KeyError:
                # XXX current Python does not support add title and description into mutually exclusive group
                #   so the message in ex_group is dropped.
                mu_ex_groups[arg.ex_group] = tp = ap.add_mutually_exclusive_group()

            if arg.required:
                tp.required = True
        else:
            tp = ap

        arg.add_argument(tp, instance)

    # setup grouped arguments
    for group, args in groups.items():
        pp = ap.add_argument_group(group)
        mu_ex_groups: dict[str, argparse._ActionsContainer] = {}

        for arg in args:
            if arg.ex_group is not None:
                try:
                    tp = mu_ex_groups[arg.ex_group]
                except KeyError:
                    mu_ex_groups[arg.ex_group] = tp = pp.add_mutually_exclusive_group()

                if arg.required:
                    tp.required = True
            else:
                tp = pp

            arg.add_argument(tp, instance)

    return ap


def set_options(instance: T, result: argparse.Namespace) -> T:
    """set argument to ``instance``'s attributes from ``argparse.Namespace`` .

    :param instance: any instance that contains ``argument``.
    :param result:
    :return: *instance* itself.
    """
    for arg in foreach_arguments(instance):
        try:
            value = getattr(result, arg.attr)
        except AttributeError:
            pass
        else:
            arg.__set__(instance, value)

    from .commands import get_sub_command_group
    if (sub := get_sub_command_group(instance)) is not None:
        try:
            value = getattr(result, sub.attr)
        except AttributeError:
            sub.__set__(instance, None)
        else:
            sub.__set__(instance, value)

    return instance


def parse_args(instance: T, args: list[str] = None) -> T:
    """Parse the command-list arguments and apply the parsed values to the given instance

    :param instance: any instance that contains ``argument``.
    :param args: A list of strings representing command-line arguments. If ``None``, uses ``sys.argv``
    :return: ``instance`` itself, with attributes populated
    """
    ap = new_parser(instance, reset=True)
    ot = ap.parse_args(args)
    return set_options(instance, ot)


@overload
def print_help(instance, file: TextIO = sys.stdout, prog: str = None):
    pass


@overload
def print_help(instance, file: Literal[None], prog: str = None) -> str:
    pass


def print_help(instance, file: TextIO = sys.stdout, prog: str = None):
    """
    print help document.

    :param instance: any instance that contains ``argument``.
    :param file: output stream.
    :param prog: program name.
    :return: help document string if ``file`` is ``None``. Otherwise, nothing return.
    """
    buf = None
    if file is None:
        import io
        buf = file = io.StringIO()

    if not isinstance(instance, ArgumentParser):
        instance = new_parser(instance, prog=prog)

    instance.print_help(file)
    if buf is not None:
        return buf.getvalue()
    else:
        return None


def with_defaults(instance: T) -> T:
    """Initialize all argument attributes by assign the default value if provided.

    :param instance: any instance that contains ``argument``.
    :return: *instance* itself, with attributes initialized with proper default.
    """
    for arg in foreach_arguments(instance):
        try:
            value = arg.default
        except ValueError:
            arg.__delete__(instance)
        else:
            arg.__set__(instance, value)

    from .commands import get_sub_command_group
    if (sub := get_sub_command_group(instance)) is not None:
        sub.__set__(instance, None)

    return instance


def as_dict(instance: T) -> dict[str, Any]:
    """collect all argument attributes into a dictionary with attribute name to its value.

    :param instance: any instance that contains ``argument``.
    :return: A dictionary mapping argument attribute names to their values
    """
    ret = {}
    for arg in foreach_arguments(instance):
        try:
            value = arg.__get__(instance)
        except AttributeError:
            pass
        else:
            ret[arg.attr] = value

    from .commands import get_sub_command_group
    if (sub := get_sub_command_group(instance)) is not None:
        try:
            value = sub.__get__(instance)
        except AttributeError:
            pass
        else:
            ret[sub.attr] = value

    return ret


def copy_argument(opt: T, ref, **kwargs) -> T:
    """copy argument from ref to opt

    :param opt: any instance that contains ``argument``.
    :param ref: any instance that contains ``argument``.
    :param kwargs: overwrite argument value mapping.
    :return: ``opt`` itself.
    """
    shadow = ShadowOption(ref, **kwargs)

    for arg in foreach_arguments(opt):
        try:
            value = getattr(shadow, arg.attr)
        except AttributeError:
            pass
        else:
            # print('set', arg.attr, value)
            arg.__set__(opt, value)

    from .commands import get_sub_command_group
    if (sub := get_sub_command_group(opt)) is not None:
        try:
            value = getattr(shadow, opt)
        except AttributeError:
            pass
        else:
            sub.__set__(opt, value)

    return opt


class ShadowOption:
    """
    (internal) Do not use class directly.

    Shadow options, used to pass wrapped :class:`AbstractOptions`
    """

    def __init__(self, ref, **kwargs):
        self.__ref = ref
        self.__kwargs = kwargs

    def __getattr__(self, attr: str):
        if attr in self.__kwargs:
            return self.__kwargs[attr]

        if attr.startswith('_') and attr[1:] in self.__kwargs:
            return self.__kwargs[attr[1:]]

        if hasattr(self.__ref, attr):
            return getattr(self.__ref, attr)

        raise AttributeError(attr)



from __future__ import annotations

from typing import TYPE_CHECKING

from .core import ArgumentParser, copy_argument

if TYPE_CHECKING:
    import polars as pl

__all__ = ['Cloneable']


class Cloneable:
    def __init__(self, ref: ArgumentParser | Cloneable | pl.DataFrame | None = None, /, **kwargs):
        if ref is not None:
            _copy_argument(self, ref, kwargs)


def _copy_argument(self: Cloneable, ref: Cloneable | pl.DataFrame, kwargs: dict):
    try:
        import polars as pl
    except ImportError:
        copy_argument(self, ref, **kwargs)
    else:
        if isinstance(ref, pl.DataFrame):
            _copy_argument_polars_dataframe(self, ref, kwargs)
        else:
            copy_argument(self, ref, **kwargs)


def _copy_argument_polars_dataframe(self: Cloneable, ref: pl.DataFrame, kwargs: dict):
    dataset = ref.unique()
    if len(dataset) > 1:
        raise RuntimeError(f'dataset not unique : {ref}')

    for column in dataset.columns:
        kwargs.setdefault(column, dataset[0, column])

    copy_argument(self, None, **kwargs)



from __future__ import annotations

import inspect
import textwrap
from collections.abc import Callable
from typing import NamedTuple, TypeVar, Any

from typing_extensions import Self

__all__ = [
    'DispatchCommand',
    'DispatchCommandNotFound',
    'Dispatch',
]

T = TypeVar('T')
R = TypeVar('R')

ARGCLZ_DISPATCH_GROUP = '__argclz_dispatch_group__'
ARGCLZ_DISPATCH_COMMAND = '__argclz_dispatch_command__'


class DispatchCommand(NamedTuple):
    host: type[T] | None
    group: str | None
    command: str
    aliases: tuple[str, ...]
    order: float
    usage: str | None
    func: Callable[..., R]  # target function
    validators: dict[str, Callable[[str], Any]]
    hidden: bool = False

    @property
    def commands(self) -> list[str]:
        return [self.command, *self.aliases]

    def parameters(self) -> list[CommandParameter]:
        s = inspect.signature(self.func)
        p = inspect.Parameter
        return [CommandParameter.of(name, para) for i, (name, para) in enumerate(s.parameters.items()) if i > 0]

    @property
    def doc(self) -> str | None:
        return self.func.__doc__

    def __call__(self, zelf: T, *args, **kwargs) -> R:
        _args = []
        _kwargs = dict(kwargs)

        for value in args:
            if isinstance(value, str):
                match value.partition('='):
                    case (value, '', ''):
                        _args.append(value)
                    case (k, '=', value):
                        _kwargs[k] = value
            else:
                _args.append(value)

        a = inspect.signature(self.func).bind_partial(zelf, *_args, **_kwargs)

        for par, validator in self.validators.items():
            try:
                val = a.arguments[par]
            except KeyError:
                continue

            try:
                new_value = validator(val)
            except (ValueError, TypeError, IndexError, KeyError) as e:
                raise ValueError(f'command {self.command} argument "{par}" : {e}') from e

            a.arguments[par] = new_value

        return self.func(*a.args, **a.kwargs)


class CommandParameter(NamedTuple):
    name: str
    optional: bool
    kind: inspect._ParameterKind

    @classmethod
    def of(cls, name: str, para: inspect.Parameter) -> Self:
        optional = para.default is not inspect.Parameter.empty
        return CommandParameter(name, optional, para.kind)

    def usage(self):
        name = self.name.upper()
        match self.kind:
            case inspect.Parameter.VAR_KEYWORD:
                return f'**{name}'
            case inspect.Parameter.VAR_POSITIONAL:
                return f'*{name}'
            case inspect.Parameter.KEYWORD_ONLY:
                ret = f'{name}='
            case _:
                ret = name

        if self.optional:
            return f'[{ret}]'
        else:
            return ret


class DispatchCommandNotFound(RuntimeError):
    def __init__(self, command: str, group: str | None = None):
        if group is None:
            message = f'command {command} not found'
        else:
            message = f'command {group}:{command} not found'

        super().__init__(message)


class CommandHelps(NamedTuple):
    commands: list[str]
    order: float
    usage: str | None
    params: list[CommandParameter]
    doc: str

    @classmethod
    def of(cls, command: DispatchCommand) -> Self:
        return CommandHelps(command.commands, command.order, command.usage, command.parameters(), command.doc or '')

    def build_command_usage(self, show_para: bool = False) -> str:
        if self.usage is not None:
            return self.usage

        match self.commands:
            case [command]:
                ret = command
            case [command, *aliases]:
                ret = command + ' (' + ', '.join(aliases) + ')'
            case _:
                raise RuntimeError()

        if not show_para:
            return ret

        return ret + ' ' + ' '.join([it.usage() for it in self.params])

    def brief_doc(self) -> str:
        contents = textwrap.dedent(self.doc).split('\n')

        ret = []
        for content in contents:
            content = content.strip()
            if content == '':
                if len(ret):
                    break
                else:
                    continue

            if content.endswith('.'):
                ret.append(content)
                break

            try:
                i = content.index('. ')
            except ValueError:
                pass
            else:
                ret.append(content[:i + 1])
                break

            ret.append(content)

        return ' '.join(ret)


class Dispatch:
    @classmethod
    def list_commands(cls, group: str | None = ..., *, all: bool = False) -> list[DispatchCommand]:
        """list all dispatch-decorated function info in *host*.

        :param group: dispatch group.
        :param all: including hidden commands
        :return: list of DispatchCommand
        """
        info: DispatchCommand

        ret = []
        for attr in dir(cls):
            attr_value = getattr(cls, attr)
            if (info := getattr(attr_value, ARGCLZ_DISPATCH_COMMAND, None)) is not None:
                if group is ... or group == info.group:
                    if all or not info.hidden:
                        ret.append(info)

        return ret

    @classmethod
    def find_command(cls, command: str, group: str | None = ...) -> DispatchCommand | None:
        """find dispatch-decoratored function in *host* according to *command*.

        :param command: command or command alias
        :param group: dispatch group
        :return: found DispatchCommand
        """
        info: DispatchCommand

        for attr in dir(cls):
            attr_value = getattr(cls, attr)
            if (info := getattr(attr_value, ARGCLZ_DISPATCH_COMMAND, None)) is not None:
                if group is ... or group == info.group:
                    if command == info.command or command in info.aliases:
                        return info

        return None

    def invoke_command(self, command: str, *args, **kwargs) -> Any:
        """invoke a dispatch-decoratored function in default group.

        :param command: command or command alias
        :param args: dispatch-decoratored function positional arguments
        :param kwargs: dispatch-decoratored function keyword arguments
        :return: function return
        :raise DispatchCommandNotFound:
        """
        if (info := self.find_command(command, None)) is None:
            raise DispatchCommandNotFound(command)
        return info(self, *args, **kwargs)

    def invoke_group_command(self, group: str, command: str, *args, **kwargs) -> Any:
        """invoke a dispatch-decoratored function in certain group.

        :param group: dispatch group
        :param command: command or command alias
        :param args: dispatch-decoratored function positional arguments
        :param kwargs: dispatch-decoratored function keyword arguments
        :return: function return
        :raise DispatchCommandNotFound:
        """
        if (info := self.find_command(command, group)) is None:
            raise DispatchCommandNotFound(command, group)
        return info(self, *args, **kwargs)

    @classmethod
    def build_command_usages(cls, group: str | None = None, *,
                             show_para: bool = False,
                             width: int = 120,
                             doc_indent: int = 20) -> str:
        ret = []

        commands = cls.list_commands(group)
        commands.sort(key=lambda it: it.order)

        for info in commands:
            info = CommandHelps.of(info)

            header = info.build_command_usage(show_para=show_para)
            content = info.brief_doc()

            if len(header) < doc_indent:
                content = header + ' ' * (doc_indent - len(header)) + content
                ret.extend(textwrap.wrap(content, width,
                                         subsequent_indent=' ' * doc_indent,
                                         break_long_words=True,
                                         break_on_hyphens=True))
            else:
                ret.append(header)
                ret.extend(textwrap.wrap(content, width,
                                         initial_indent=' ' * doc_indent,
                                         subsequent_indent=' ' * doc_indent,
                                         break_long_words=True,
                                         break_on_hyphens=True))

        return '\n'.join(ret)
